/**********************************************************

This file is part of my 3d engine (DXC) and is not designed 
to be used in other applications. It is neither optimized 
nor does it cover all features of RTX. I can not guarantee
is functionality and I am not responsible for any malfunctions 
including possible damage to your computer system. Yet, I
hope it may be helpful to use parts of the code to load RTX 
files to your application.
This source code my be distributed and used freely.

                           Michael.Schiestl@cluetec.com

**********************************************************/

RTXLoader::RTXLoader() : DXC_Base()
{
	SetCaller("RTXLoader"); Set0();
}

RTXLoader::~RTXLoader()
{
	
}

void RTXLoader::DXCWarn(char *err, ...)
{
	ERRSTR er,er2;
	
	va_list parameter;             
	va_start( parameter, err );
    vsprintf(er2, err, parameter );
			
	strcpy(er,FileName);
	strcat(er,": ");
	strcat(er,er2);
	DXC_Base::DXCWarn(er);
}

void RTXLoader::DXCError(char *err, ...)
{
	ERRSTR er,er2;
	
	va_list parameter;             
	va_start( parameter, err );
    vsprintf( er2, err, parameter );
			
	strcpy(er,FileName);
	strcat(er,": ");
	strcat(er,er2);
	DXC_Base::DXCError(er);
}

void RTXLoader::Set0()
{
	m_fp=NULL; vbuf=NULL; fbuf=NULL; tvbuf=NULL; tfbuf=NULL; filebuf=NULL; dgfx=NULL;
	strcpy(FileName,"*nofile*"); m_LoadTex=TRUE;
}

int RTXLoader::Str2Id(char*str)
{
	char st[128];
	strcpy(st,str);
	_strupr(st);
	if(strcmp(st,"SCENE")==0) return RTX_SCENE;
	if(strcmp(st,"MATERIAL_LIST")==0) return RTX_MATERIAL_LIST;
	if(strcmp(st,"GEOMOBJECT")==0) return RTX_GEOMOBJECT;
	if(strcmp(st,"SCENE_BACKGROUND_STATIC")==0) return RTX_BACKGROUND_STATIC;
	if(strcmp(st,"SCENE_AMBIENT_STATIC")==0) return RTX_AMBIENT_STATIC;
	if(strcmp(st,"SCENE_FIRSTFRAME")==0) return RTX_FIRSTFRAME;
	if(strcmp(st,"SCENE_LASTFRAME")==0) return RTX_LASTFRAME;
	if(strcmp(st,"SCENE_FRAMESPEED")==0) return RTX_FRAMESPEED;
	if(strcmp(st,"SCENE_TICKSPERFRAME")==0) return RTX_TICKSPERFRAME;
	if(strcmp(st,"SCENE_ENVMAP")==0) return RTX_SCENE_ENVMAP;
	if(strcmp(st,"MATERIAL_COUNT")==0) return RTX_MATERIAL_COUNT;
	if(strcmp(st,"MATERIAL")==0) return RTX_MATERIAL;
	if(strcmp(st,"MATERIAL_NAME")==0) return RTX_MATERIAL_NAME;
	if(strcmp(st,"MAP_DIFFUSE")==0) return RTX_MAP_DIFFUSE;
	if(strcmp(st,"UVW_U_OFFSET")==0) return RTX_MAP_U_OFF;
	if(strcmp(st,"UVW_V_OFFSET")==0) return RTX_MAP_V_OFF;
	if(strcmp(st,"UVW_U_TILING")==0) return RTX_MAP_U_TILE;
	if(strcmp(st,"UVW_V_TILING")==0) return RTX_MAP_V_TILE;
	if(strcmp(st,"UVW_U_ANGLE")==0) return RTX_MAP_U_ANGLE;
	if(strcmp(st,"UVW_V_ANGLE")==0) return RTX_MAP_V_ANGLE;
	if(strcmp(st,"UVW_W_ANGLE")==0) return RTX_MAP_W_ANGLE;
	if(strcmp(st,"BITMAP")==0) return RTX_BITMAP;
	if(strcmp(st,"NODE_NAME")==0) return RTX_NODE_NAME;
	if(strcmp(st,"NODE_VISIBLE")==0) return RTX_NODE_VISIBLE;
	if(strcmp(st,"NODE_PARENT")==0) return RTX_NODE_PARENT;
	if(strcmp(st,"MESH")==0) return RTX_MESH;
	if(strcmp(st,"MESH_NUMVERTEX")==0) return RTX_MESH_NUMVERTEX;
	if(strcmp(st,"MESH_NUMFACES")==0) return RTX_MESH_NUMFACES;
	if(strcmp(st,"MESH_VERTEX_LIST")==0) return RTX_MESH_VERTEX_LIST;
	if(strcmp(st,"MESH_VERTEX")==0) return RTX_MESH_VERTEX;
	if(strcmp(st,"MESH_FACE_LIST")==0) return RTX_MESH_FACE_LIST;
	if(strcmp(st,"MESH_FACE")==0) return RTX_MESH_FACE;
	if(strcmp(st,"MESH_NUMTVERTEX")==0) return RTX_MESH_NUMTVERTEX;
	if(strcmp(st,"MESH_TVERTLIST")==0) return RTX_MESH_TVERTLIST;
	if(strcmp(st,"MESH_TVERT")==0) return RTX_MESH_TVERT;
	if(strcmp(st,"MESH_NUMTVFACES")==0) return RTX_MESH_NUMTVFACES;
	if(strcmp(st,"MESH_TFACELIST")==0) return RTX_MESH_TFACELIST;
	if(strcmp(st,"MESH_TFACE")==0) return RTX_MESH_TFACE;
	if(strcmp(st,"MESH_SMOOTHING")==0) return RTX_MESH_SMOOTHING;
	if(strcmp(st,"MESH_MTLID")==0) return RTX_MESH_MTLID;
	if(strcmp(st,"MESH_WEIGHTS")==0) return RTX_MESH_WEIGHTS;
	if(strcmp(st,"MATERIAL_REF")==0) return RTX_MATERIAL_REF;
	if(strcmp(st,"MATERIAL_DIFFUSE")==0) return RTX_MATERIAL_DIFFUSE;
	if(strcmp(st,"MATERIAL_AMBIENT")==0) return RTX_MATERIAL_AMBIENT;
	if(strcmp(st,"MATERIAL_SPECULAR")==0) return RTX_MATERIAL_SPECULAR;
	if(strcmp(st,"MATERIAL_SHINESTRENGTH")==0) return RTX_MATERIAL_SHINE;
	if(strcmp(st,"MATERIAL_SHINE")==0) return RTX_MAT_GLOSSINESS;
	if(strcmp(st,"MATERIAL_TRANSPARENCY")==0) return RTX_MAT_TRANSPARENCY;
	if(strcmp(st,"MATERIAL_TWOSIDED")==0) return RTX_MAT_2SIDED;
	if(strcmp(st,"NUMSUBMTLS")==0) return RTX_MAT_NUMSUB;
	if(strcmp(st,"SUBMATERIAL")==0) return RTX_MAT_SUB;
	if(strcmp(st,"GROUP")==0) return RTX_GROUP;
	if(strcmp(st,"TM_ANIMATION")==0) return RTX_ANIMATION;
	if(strcmp(st,"CONTROL_POS_BEZIER")==0) return RTX_CTRL_POS;
	if(strcmp(st,"CONTROL_BEZIER_POS_KEY")==0) return RTX_CTRL_POS_KEY_BEZ;
	if(strcmp(st,"CONTROL_ROT_TCB")==0) return RTX_CTRL_ROT;
	if(strcmp(st,"CONTROL_TCB_ROT_KEY")==0) return RTX_CTRL_ROT_KEY_TCB;
	if(strcmp(st,"CONTROL_SCALE_BEZIER")==0) return RTX_CTRL_SCALE;
	if(strcmp(st,"CONTROL_BEZIER_SCALE_KEY")==0) return RTX_CTRL_SCALE_KEY_BEZ;
	if(strcmp(st,"CONTROL_POS_TRACK")==0) return RTX_CTRL_POS;
	if(strcmp(st,"CONTROL_ROT_TRACK")==0) return RTX_CTRL_ROT;
	if(strcmp(st,"CONTROL_SCALE_TRACK")==0) return RTX_CTRL_SCALE;
	if(strcmp(st,"CONTROL_POS_SAMPLE")==0) return RTX_CTRL_POS_KEY;
	if(strcmp(st,"CONTROL_ROT_SAMPLE")==0) return RTX_CTRL_ROT_KEY;
	if(strcmp(st,"CONTROL_SCALE_SAMPLE")==0) return RTX_CTRL_SCALE_KEY;
	if(strcmp(st,"CAMERAOBJECT")==0) return RTX_CAMERA; 
	if(strcmp(st,"CAMERA_SETTINGS")==0) return RTX_CAM_SETTINGS;
	if(strcmp(st,"CAMERA_FOV")==0) return RTX_CAM_FOV;
	if(strcmp(st,"NODE_TM")==0) return RTX_NODE_TM;
	if(strcmp(st,"TM_ROW0")==0) return RTX_TM_ROW0;
	if(strcmp(st,"TM_ROW1")==0) return RTX_TM_ROW1;
	if(strcmp(st,"TM_ROW2")==0) return RTX_TM_ROW2;
	if(strcmp(st,"TM_ROW3")==0) return RTX_TM_ROW3;
	if(strcmp(st,"BONEOBJECT")==0) return RTX_BONE;
	if(strcmp(st,"HELPEROBJECT")==0) return RTX_HELPER;
	if(strcmp(st,"TM_POS")==0) return RTX_TM_POS;
	if(strcmp(st,"TM_SCALE")==0) return RTX_TM_SCALE;
	if(strcmp(st,"TM_ROTAXIS")==0) return RTX_TM_ROTAXIS;
	if(strcmp(st,"TM_ROTANGLE")==0) return RTX_TM_ROTANGLE;
	if(strcmp(st,"TARGET_NODE")==0) return RTX_CAM_TARGET;
	if(strcmp(st,"TARGETOBJECT")==0) return RTX_TARGETOBJ;
	if(strcmp(st,"LIGHTOBJECT")==0) return RTX_LIGHT;
	if(strcmp(st,"LIGHT_SETTINGS")==0) return RTX_LIGHT_SETTINGS;
	if(strcmp(st,"LIGHT_COLOR")==0) return RTX_LIGHT_COLOR;
	if(strcmp(st,"LIGHT_INTENS")==0) return RTX_LIGHT_INTENS;
	if(strcmp(st,"LIGHT_ASPECT")==0) return RTX_LIGHT_ASPECT;
	if(strcmp(st,"LIGHT_HOTSPOT")==0) return RTX_LIGHT_HOTSPOT;
	if(strcmp(st,"LIGHT_FALLOFF")==0) return RTX_LIGHT_FALLOFF;
	if(strcmp(st,"LIGHT_TDIST")==0) return RTX_LIGHT_TDIST;
	if(strcmp(st,"LIGHT_USELIGHT")==0) return RTX_USELIGHT;
	if(strcmp(st,"LIGHT_TYPE")==0) return RTX_LIGHT_TYPE;
	if(strcmp(st,"MESH_INIT_TM")==0) return RTX_MESH_INIT_TM;
	if(strcmp(st,"BONE_WEIGHT")==0) return RTX_BONE_WEIGHT;
	if(strcmp(st,"BONE_NAME")==0) return RTX_BONE_NAME;
	if(strcmp(st,"BONE_INIT_TM")==0) return RTX_BONE_INIT_TM;
	if(strcmp(st,"INDEX_WEIGHT")==0) return RTX_INDEX_WEIGHT;
	//if(strcmp(st,"")==0) return RTX_
	return 0;
}


int RTXLoader::absmax3(float v1, float v2, float v3)
{
	float w1,w2,w3;
	w1=(float)fabs(v1); w2=(float)fabs(v2); w3=(float)fabs(v3);
	if(w1>w2) if(w1>w3) return 0;
			  else return 2;
	else if(w2>w3) return 1;
		 else return 2;
}

int RTXLoader::vread(void *ptr,int size)
{
	if(size+m_curFilePos>m_FileSize) size=m_FileSize-m_curFilePos;
	if(m_bBufRead) {
		memcpy(ptr,&filebuf[m_curFilePos],size);
	} else size=fread(ptr,1,size,m_fp);
	m_curFilePos+=size;
	return size;
}

BOOL RTXLoader::vseek(int rel)
{
	BOOL b=TRUE;
	int setpos;
	if(m_curFilePos+rel<0) { setpos=0; b=FALSE; }
	else if(m_curFilePos+rel>m_FileSize) { setpos=m_FileSize; b=FALSE; }
		else setpos=m_curFilePos+rel;
	if(!m_bBufRead) fseek(m_fp,setpos,SEEK_SET);
	m_curFilePos=setpos;
	return b;
}

BOOL RTXLoader::Goto(char ch)
{
	char dum;
	do {
		if(vread(&dum,1)!=1) DXCError("'%c' not found",dum);
	} while(dum!=ch);
	return FALSE;
}

BOOL RTXLoader::Skip()
{
	char dum;
	do {
		if(vread(&dum,1)!=1) return TRUE;
		
	} while(dum<=32);
	vseek(-1);
	return FALSE;
}

int RTXLoader::ReadNext()
{
	char dum,str[128];
	int sub=0;
	do {
		if(vread(&dum,1)!=1) { 
			return -1;
		}
		if(dum=='{') sub++;
		if(dum=='}') {
			if(sub>0) sub--;
			else return -2;
		}
	} while(dum!='*'||sub!=0);
	ReadText2(str);
	return Str2Id(str);
}

char RTXLoader::ReadChar()
{
	char ch;
	vread(&ch,1);
	return ch;
}

char RTXLoader::NextChar()
{
	char ch;
	vread(&ch,1); vseek(-1); 
	return ch;
}

BOOL RTXLoader::ReadInt(int *f)
{
	char buf[128];
	double d;
	BOOL b;
	Skip();
	b=ReadNumber(buf);
	d=strtod(buf,NULL);
	*f=(int)d;
	return b;
}

BOOL RTXLoader::ReadFloat(float *f)
{
	char buf[128];
	double d;
	BOOL b;
	Skip();
	b=ReadText2(buf);
	d=strtod(buf,NULL);
	*f=(float)d;
	return b;
}

BOOL RTXLoader::ReadFVect3D(FVect3D *v)
{
	if(ReadFloat(&v->x)) return TRUE;
	if(ReadFloat(&v->y)) return TRUE;
	if(ReadFloat(&v->z)) return TRUE;
	return FALSE;
}

BOOL RTXLoader::ReadFVect2D(FVect2D *v)
{
	if(ReadFloat(&v->x)) return TRUE;
	if(ReadFloat(&v->y)) return TRUE;
	return FALSE;
}

BOOL RTXLoader::ReadNumber(char *txt)
{
	char dum;
	int i=0;
	do {
		if(vread(&dum,1)!=1) return TRUE;
		txt[i]=dum; i++; if(i==128) DXCError("String too long (>127)");
	} while( (dum>=48&&dum<58) || dum=='-' || dum=='+' || dum=='.');
	txt[i-1]=0;
	vseek(-1);
	if(i==1) return TRUE; // error
	return FALSE; // ok
}

BOOL RTXLoader::ReadText2(char *txt)
{
	char dum;
	int i=0;
	do {
		if(vread(&dum,1)!=1) return TRUE;
		txt[i]=dum; i++; if(i==128) DXCError("String too long (>127)");
	} while(dum>32&&dum!='{'&&dum!='}'&&dum!='"');
	txt[i-1]=0;
	vseek(-1);
	return FALSE;
}

BOOL RTXLoader::ReadText(char *txt)
{
	char dum;
	int i=0;
	Goto('"');
	do {
		if(vread(&dum,1)!=1) return TRUE;
		txt[i]=dum; i++; if(i==128) DXCError("String too long (>127)");
	} while(dum!='"');
	txt[i-1]=0;
	return FALSE;
}

BOOL RTXLoader::ReadFaceList()
{
	int id,dum,i,nr=-1;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_MESH_SMOOTHING: Skip(); 
			i=0;
			if(NextChar()!='*')
			for(;;) {
				ReadInt(&dum);
				fbuf[nr].sg.SmGr[i]=(BYTE)dum; 
				Skip(); i++;
				if(NextChar()!=',') break;
				ReadChar();
				Skip();
			}
			fbuf[nr].sg.NrSmGr=i;
			break;
		case RTX_MESH_MTLID: ReadInt(&dum); fbuf[nr].MatId=dum;	break;
		case RTX_MESH_FACE: ReadInt(&dum); nr++; if(dum!=nr) DXCError("Wrong face list order!");
			if(nr>=numFaces) DXCError("Too many faces");
			Goto('A'); Goto(':'); ReadInt(&dum); fbuf[nr].v[0]=dum;
			Goto('B'); Goto(':'); ReadInt(&dum); fbuf[nr].v[1]=dum;
			Goto('C'); Goto(':'); ReadInt(&dum); fbuf[nr].v[2]=dum;
			break;
		}
	}
}

BOOL RTXLoader::ReadTFaceList()
{
	int id,dum,nr=0;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_MESH_TFACE: ReadInt(&dum); if(dum!=nr) DXCError("Wrong face list order!");
			if(nr>=numTFaces) DXCError("Too many t-faces");
			ReadInt(&dum); tfbuf[nr].v[0]=dum;
			ReadInt(&dum); tfbuf[nr].v[1]=dum;
			ReadInt(&dum); tfbuf[nr].v[2]=dum;
			nr++;
			break;
		}
	}
}

BOOL RTXLoader::ReadVertexList()
{
	int id,dum,nr=0;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_MESH_VERTEX: ReadInt(&dum); if(dum!=nr) DXCError("Wrong vertex list order!");
			if(nr>=numVer) DXCError("Too many vertices");
			ReadFVect3D(&vbuf[nr].pos); 
			nr++; break;
		}
	}
}

BOOL RTXLoader::ReadTVertexList()
{
	int id,dum,nr=0;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_MESH_TVERT: ReadInt(&dum); if(dum!=nr) DXCError("Wrong vertex list order!");
			if(nr>=numTVer) DXCError("Too many t-vertices");
			ReadFVect3D(&tvbuf[nr]); tvbuf[nr].y=1-tvbuf[nr].y;
			nr++; break;
		}
	}
}

// Creates the mesh. To support smoothing groups you may have to duplicate
// vertices.
void RTXLoader::CalcMesh() 
{
	int i,j,k,dum,nrSub,nrMat,numInval=0;
	DWORD vformat;
	float sum,mag;
	FVect3D normal,v1,v2;
	DXC_BoneInfo bi;
	switch(aktMesh->nrWeights) {
	case 0: vformat=DXC_VERTEX; break;
	case 1: 
		DXCWarn("Only one bone in object %s",aktMesh->Name); 
		vformat=DXC_SKINVERTEX1; aktMesh->nrWeights=2; 
		break;
	case 2: vformat=DXC_SKINVERTEX1; break;
	case 3: vformat=DXC_SKINVERTEX2; break;
	case 4: vformat=DXC_SKINVERTEX3; break;
	default: DXCError("Invalid number of weightings: %i",aktMesh->nrWeights);
	}
	aktMesh->SetVerBlendMode();
	
	if(aktMesh->pBone.GetSize()>0)
	if(aktMesh->nrWeights>0) 
	for(i=0;i<numVer;i++) 
	{
		sum=0;
		for(j=0;j<aktMesh->nrWeights;j++) {
			sum+=vbuf[i].w[j];
		}
		if(sum==0) {
			vbuf[i].w[0]=1;
			vbuf[i].dwBone=0;
		}
		else
		for(j=0;j<aktMesh->nrWeights;j++) {
			vbuf[i].w[j]/=sum;
		}
	}
	
	visize=numFaces*3; aktVer2=0;
	vbuf2=new DXC_VertexIntern2[visize];
	ZeroMemory(vbuf2,sizeof(DXC_VertexIntern2)*visize);
	if(aktMatID==-1) nrSub=0;
	else nrSub=MatList[aktMatID].numSubMat;
	if(aktMatID>=0)
		if(nrSub<=0) nrMat=1;
		else nrMat=nrSub;
	else nrMat=1;
	for(i=0;i<numFaces;i++) {
		if(aktMatID==-1||MatList[aktMatID].numSubMat==0) fbuf[i].MatId=0; // Bug in ASE ??
		v1=vbuf[fbuf[i].v[1]].pos-vbuf[fbuf[i].v[0]].pos;
		v2=vbuf[fbuf[i].v[2]].pos-vbuf[fbuf[i].v[0]].pos;
		normal=v1.Cross(v2); mag=normal.Mag();
		if(mag==0) {
			DXCWarn("Invalid face %i in mesh '%s'",i,aktMesh->Name);
			fbuf[i].Attr=-1; numInval++;
			continue;
		} else normal/=mag;
		for(j=0;j<3;j++) {
			//make sure that faces from different smoothing groups or submaterial get their own vertices
			dum=RequestVertex(fbuf[i].v[j],&fbuf[i].sg,normal,fbuf[i].MatId);
			fbuf[i].w[j]=dum;
			if(aktMesh->IsTexCoords) {
				if(vbuf2[dum].numTex>0) {
					if(vbuf2[dum].numTex==1) {
						if( vbuf2[dum].t.x!=(tvbuf[tfbuf[i].v[j]].x) ||
							vbuf2[dum].t.y!=(tvbuf[tfbuf[i].v[j]].y) ) {
							vbuf2[dum].t.x+=(tvbuf[tfbuf[i].v[j]].x);
							vbuf2[dum].t.y+=(tvbuf[tfbuf[i].v[j]].y);	
							vbuf2[dum].numTex++; // average the mapping coords if multiple
							DXCWarn("Multiple mapping coords in mesh '%s'",aktMesh->Name);
						}
					}
				} else {
					vbuf2[dum].t.x=(tvbuf[tfbuf[i].v[j]].x);
					vbuf2[dum].t.y=(tvbuf[tfbuf[i].v[j]].y);
					vbuf2[dum].numTex++; 
				}
			}
		}
		if(nrSub>0) {
			fbuf[i].Attr=fbuf[i].MatId;
		} else fbuf[i].Attr=0;
	}
	
	// aktMesh will be your final mesh
	dgfx->CreateMesh(aktMesh,numFaces-numInval,aktVer2,nrMat,vformat);
	
	if(nrSub>0)
	for(i=0;i<nrSub;i++)
	{	if(aktMatID>=0) {
			aktMesh->SetMaterial(&MatList[aktMatID].SubMat[i].Mat,i);
			aktMesh->SetTexture(&MatList[aktMatID].SubMat[i].Tex,i);
		}
	} else
		if(aktMatID>=0) {
			aktMesh->SetMaterial(&MatList[aktMatID].Mat,0);
			aktMesh->SetTexture(&MatList[aktMatID].Tex,0);
		}
	
	aktMesh->Lock();
	for(i=0,k=0;i<numFaces;i++) {
		if(fbuf[i].Attr!=-1) {
			for(j=0;j<3;j++) {
				aktMesh->Face(k)->v[j]=fbuf[i].w[j];
			}
			*aktMesh->Attribute(k)=fbuf[i].Attr;
			k++;
		}
	}
	for(i=0;i<aktVer2;i++) {
		*aktMesh->VertexP(i)=vbuf2[i].pos;
		if(vbuf2[i].numTex>0)
			*aktMesh->VertexT(i)=vbuf2[i].t/vbuf2[i].numTex; //the texture coodrs of your mesh
		else *aktMesh->VertexT(i)=FVect2D(0,0);
		*aktMesh->VertexN(i)=vbuf2[i].NormalSum/vbuf2[i].NrFaces;
		mag=aktMesh->VertexN(i)->Mag();
		if(mag==0) DXCWarn("Zero normal sum in object %s!",aktMesh->Name);
		*aktMesh->VertexN(i)/=mag; //the normals of your mesh
		if(aktMesh->nrWeights>0) {
			*aktMesh->VertexM(i)=vbuf2[i].dwBone; //the matrix indices of your mesh
			for(j=0;j<aktMesh->nrWeights-1;j++) {
				aktMesh->VertexW(i)[j]=vbuf2[i].w[j]; //the weightings of your mesh
			}
		}
	}
	aktMesh->Unlock();
	aktMesh->Optimize();
	DXC_Delete(vbuf2); visize=0;
}

int RTXLoader::IntersectSmGr(DXC_SmoothingGroupSet *sg_in1, DXC_SmoothingGroupSet *sg_in2, DXC_SmoothingGroupSet *sg_out)
{
	int i,j;
	sg_out->NrSmGr=0;
	for(i=0;i<sg_in1->NrSmGr;i++) 
		for(j=0;j<sg_in2->NrSmGr;j++) {
			if(sg_in1->SmGr[i]==sg_in2->SmGr[j]) {
				sg_out->SmGr[sg_out->NrSmGr]=sg_in2->SmGr[j];
				sg_out->NrSmGr++;
			}
		}
	return sg_out->NrSmGr;
}

int RTXLoader::RequestVertex(int nr, DXC_SmoothingGroupSet *sg, FVect3D normal, int SubMat)
{
	int i,j,id;
	DXC_SmoothingGroupSet *pSg;
	DXC_SmoothingGroupSet Sg;
	for(i=0;i<vbuf[nr].NumVer;i++) {
		id=vbuf[nr].VerId[i];
		pSg=&vbuf2[id].sg;
		if(SubMat==vbuf2[id].SubMat) 
		{
			if(IntersectSmGr(sg,pSg,&Sg)) 
			{
				*pSg=Sg; vbuf2[id].NormalSum+=normal;
				vbuf2[id].NrFaces++;
				return id;
			}
		}
	}
	if(vbuf[nr].NumVer>=RTX_MAX_VERMULTI) {
		DXCWarn("Too many smoothing groups on vertex %i!",nr);
		id=vbuf[nr].VerId[0]; pSg=&vbuf2[id].sg;
		*pSg=Sg; vbuf2[id].NormalSum+=normal;
		vbuf2[id].NrFaces++;
		return id;
	}
	if(aktVer2==visize) DXCError("Vertex buffer too small!");

	vbuf[nr].VerId[vbuf[nr].NumVer]=aktVer2; vbuf[nr].NumVer++;
	vbuf2[aktVer2].pos=vbuf[nr].pos;
	for(j=0;j<4;j++) vbuf2[aktVer2].w[j]=vbuf[nr].w[j];
	vbuf2[aktVer2].dwBone=vbuf[nr].dwBone;
	vbuf2[aktVer2].sg=*sg; vbuf2[aktVer2].NormalSum=normal;
	vbuf2[aktVer2].NrFaces=1; 
	vbuf2[aktVer2].SubMat=SubMat;
	aktVer2++;
	return aktVer2-1;
}

int RTXLoader::AddToBoneList(DXC_BoneInfo *bi)
{
	int i;
	for(i=0;i<aktMesh->pBone.GetSize();i++) {
		if(strcmp(aktMesh->pBone(i).Name,bi->Name)==0) return aktMesh->pBone.List(i);
	}
	return aktMesh->pBone.Add(*bi);
}

BOOL RTXLoader::ReadBoneWeight()
{
	int id,dum,n,nr=0,m;
	DXC_TM_Info ti;
	DXC_BoneInfo bi;
	D3DXMATRIX mat;
	float f;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_BONE_NAME: ReadText(bi.Name);  break;
		case RTX_BONE_INIT_TM: Goto('{'); ReadTM(&ti); 
			D3DXMatrixInverse(&mat,NULL,&m_MeshInitTM.mat);
			bi.m_BoneInitInv=ti.mat; 
			bi.m_MeshInit=m_MeshInitTM.mat; 
			bi.m_MeshInitInv=mat; 
			nr=AddToBoneList(&bi);
			break;
		case RTX_INDEX_WEIGHT: ReadInt(&dum); ReadFloat(&f);
			if(dum>=numVer) DXCError("Invalid weight index: %i",dum); 
			n=vbuf[dum].NumBone;
			if(n>=4) { // discard the bone with the least weight
				if(vbuf[dum].w[0]<vbuf[dum].w[1]) {
					if(vbuf[dum].w[2]<vbuf[dum].w[3]) {
						if(vbuf[dum].w[2]<vbuf[dum].w[0]) m=2;
						else m=0;
					} else {
						if(vbuf[dum].w[0]<vbuf[dum].w[3]) m=0;
						else m=3;
					}
				} else {
					if(vbuf[dum].w[2]<vbuf[dum].w[3]) {
						if(vbuf[dum].w[2]<vbuf[dum].w[1]) m=2;
						else m=1;
					} else {
						if(vbuf[dum].w[1]<vbuf[dum].w[3]) m=1;
						else m=3;
					}
				}
				vbuf[dum].w[m]=f;
				vbuf[dum].Bones[m]=nr;
				DXCWarn("More than 4 bones in vertex of object '%s'",aktMesh->Name); break; 
			}
			vbuf[dum].Bones[n]=nr; vbuf[dum].w[n]=f; n++; vbuf[dum].NumBone=n;
			if(n>aktMesh->nrWeights) aktMesh->nrWeights=n;
			break;
		}
	}
}

BOOL RTXLoader::ReadMeshWeights()
{
	int id;
	DXC_BoneInfo bi;
	D3DXMATRIX mat;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_MESH_INIT_TM: Goto('{'); ReadTM(&m_MeshInitTM); 
			D3DXMatrixIdentity(&mat);
			bi.m_BoneInitInv=mat;
			bi.m_MeshInit=mat;
			bi.m_MeshInitInv=mat;
			strcpy(bi.Name,aktMesh->Name);
			AddToBoneList(&bi);
			break;
		case RTX_BONE_WEIGHT: Goto('{'); ReadBoneWeight(); break;
		}
	}
}
	

BOOL RTXLoader::ReadMesh()
{
	int id; numFaces=0; numVer=0; numTVer=0; numTFaces=0;
	BOOL IsAdded=FALSE;
	aktMesh->IsTexCoords=FALSE;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: case -1: 
			if(id==-2) return FALSE; else return TRUE;
		case 0: break;
		case RTX_MESH_NUMVERTEX: ReadInt(&numVer); 
			if(numVer) vbuf=new DXC_VertexIntern[numVer]; 
			ZeroMemory(vbuf,sizeof(DXC_VertexIntern)*numVer); break;
		case RTX_MESH_NUMFACES: ReadInt(&numFaces); if(numFaces) fbuf=new DXC_FaceIntern[numFaces]; break;
		case RTX_MESH_NUMTVERTEX: ReadInt(&numTVer); if(numTVer) tvbuf=new FVect3D[numTVer]; break;
		case RTX_MESH_NUMTVFACES: ReadInt(&numTFaces); if(numTFaces) tfbuf=new DXC_Face[numTFaces]; break;
			if(numTFaces!=numFaces) DXCWarn("Number of t-faces and faces does not match"); break;
		case RTX_MESH_VERTEX_LIST: 
			Goto('{'); ReadVertexList();
			break;
		case RTX_MESH_FACE_LIST:
			Goto('{'); ReadFaceList(); 
			break;
		case RTX_MESH_TVERTLIST:
			Goto('{'); ReadTVertexList(); 
			break;
		case RTX_MESH_TFACELIST: 
			aktMesh->IsTexCoords=TRUE;
			Goto('{'); ReadTFaceList(); 
			break;
		case RTX_MESH_WEIGHTS:
			Goto('{'); ReadMeshWeights(); 
			aktMesh->AllocBones(); break;
		}
	}
}

BOOL RTXLoader::ReadCtrlPos()
{
	int id;
	DXC_PosKey key;
	FVect3D x0,x1,x0t,x1t;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_CTRL_POS_KEY: ReadFloat(&key.time); 
			ReadFVect3D(&key.Data.pos); aktObj->AddPosKey(&key); key.IntMode=0; break;
		case RTX_CTRL_POS_KEY_BEZ: ReadFloat(&key.time); 
			ReadFVect3D(&key.Data.pos); ReadFVect3D(&key.Data.TangIn); ReadFVect3D(&key.Data.TangOut);
			key.IntMode=1;
			aktObj->AddPosKey(&key); break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		}
	}
}

BOOL RTXLoader::ReadCtrlRot()
{
	int id;
	float lastTime=0;
	const float PI2=6.283185f/2.f;
	DXC_RotKey key;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_CTRL_ROT_KEY: key.IntMode=0; ReadFloat(&key.time); 
			if(lastTime>key.time) DXCWarn("Skipped wrong order rotation key in file: %s",FileName);
			else {
				lastTime=key.time;
				ReadFVect3D(&key.Data.axis); key.Data.axis*=DXC_ROTATION_SWAP;
				ReadFloat(&key.Data.angle); 
				if(aktObj->ObjectType==DXC_3DOBJ_BONE)
					key.Data.angle=fmodulo(key.Data.angle+PI2,2*PI2)-PI2;
				aktObj->AddRotKey(&key);
			}
			break;
		case RTX_CTRL_ROT_KEY_TCB:
			key.IntMode=2; ReadFloat(&key.time); 
			if(lastTime>key.time) DXCWarn("Skipped wrong order rotation key in file: %s",FileName);
			else {
				lastTime=key.time;
				ReadFVect3D(&key.Data.axis); key.Data.axis*=DXC_ROTATION_SWAP;
				ReadFloat(&key.Data.angle); 
				ReadFloat(&key.Data.tens); ReadFloat(&key.Data.cont); ReadFloat(&key.Data.bias);
				ReadFloat(&key.Data.easein); ReadFloat(&key.Data.easeout);
				if(aktObj->ObjectType==DXC_3DOBJ_BONE)
					key.Data.angle=fmodulo(key.Data.angle+PI2,2*PI2)-PI2;
				aktObj->AddRotKey(&key);
			}
			break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		}
	}
}

BOOL RTXLoader::ReadCtrlScale()
{
	int id; 
	DXC_ScaleKey key;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_CTRL_SCALE_KEY: ReadFloat(&key.time); /*key.time*=TimeFactor*/;
				ReadFVect3D(&key.Data.scale); key.IntMode=0;
				aktObj->AddScaleKey(&key);  break;
		case RTX_CTRL_SCALE_KEY_BEZ: ReadFloat(&key.time); /*key.time*=TimeFactor*/;
				ReadFVect3D(&key.Data.scale); key.IntMode=1;
				aktObj->AddScaleKey(&key);  break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		}
	}
}

BOOL RTXLoader::ReadAnimation()
{
	int id;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_CTRL_POS: Goto('{'); ReadCtrlPos(); break;
		case RTX_CTRL_ROT: Goto('{'); ReadCtrlRot(); break;
		case RTX_CTRL_SCALE: Goto('{'); ReadCtrlScale(); break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		}
	}
}

BOOL RTXLoader::ReadGeomObject()
{
	int id,dum,nr;
	BOOL IsMesh=FALSE;
	nr=m_Ani->pMesh.New(); aktMatID=-1;
	aktMesh=&m_Ani->pMesh[nr];
	aktMesh->ObjectType=DXC_3DOBJ_MESH;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: case -1:
			if(!IsMesh) 
			{
				dum=m_Ani->pHelper.New(); aktHelper=m_Ani->pHelper.Ptr(dum); 
				*aktHelper=*aktMesh;
				aktHelper->ObjectType=DXC_3DOBJ_HELPER;
				m_Ani->pMesh.Del(nr);
			} else {			
				CalcMesh();
				DXC_Delete(fbuf); DXC_Delete(vbuf);
				DXC_Delete(tfbuf); DXC_Delete(tvbuf);
			}
			if(id==-1) return TRUE; else return FALSE;
		case 0: break;
		case RTX_ANIMATION: Goto('{'); aktObj=aktMesh; ReadAnimation(); break;
		case RTX_NODE_NAME: ReadText(aktMesh->Name); break;
		case RTX_NODE_PARENT: ReadText(aktMesh->ParentName); break;
		case RTX_MESH: Goto('{'); ReadMesh(); IsMesh=TRUE; break;
		case RTX_MATERIAL_REF: ReadInt(&dum); aktMatID=dum; break;
		case RTX_NODE_TM: Goto('{'); ReadTM(&aktMesh->m_TM_Info);  break;
		case RTX_NODE_VISIBLE: ReadInt(&dum); 
			if(dum==0) aktMesh->IsVisible=FALSE; else aktMesh->IsVisible=TRUE;
			break;
		}
	}
}

BOOL RTXLoader::ReadEnvMap()
{
	int id;
	char txt[256];
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_BITMAP: if(ReadText(txt)) return TRUE; 
			dgfx->LoadTexture(&m_Ani->m_BgrdTex,txt);
			break;
		}
	}
}

BOOL RTXLoader::ReadMapDiffuse()
{
	int id;
	float uo,vo,ut,vt,ua,va,wa;
	FVect2D u(1,0),v(0,1);
	DXC_Matrix *m,ma;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -1: case -2: // Calculate the texture transform matrix
			m=&aktMat->matr;
			m->Identity(); ma.SetScaling(FVect3D(ut,vt,1));
			(*m)(3,0)=-0.5f; (*m)(3,1)=-0.5f;
			(*m)(3,0)-=uo; (*m)(3,1)+=vo;
			(*m).Rotate(FV3(0,0,1),-wa);
			*m*=ma;
			(*m)(3,0)+=0.5f; (*m)(3,1)+=0.5f;
			(*m)(2,0)=(*m)(3,0); (*m)(2,1)=(*m)(3,1); (*m)(2,2)=1;
			(*m)(3,0)=0; (*m)(3,1)=0; (*m)(3,2)=0; (*m)(3,3)=1;
			if(id==-2) return FALSE; else return TRUE;
		case 0: break;
		case RTX_BITMAP: if(ReadText(aktMat->DiffTextFile)) return TRUE; break;
		case RTX_MAP_U_OFF: if(ReadFloat(&uo)) return TRUE; break;
		case RTX_MAP_V_OFF: if(ReadFloat(&vo)) return TRUE; break;
		case RTX_MAP_U_TILE: if(ReadFloat(&ut)) return TRUE; break;
		case RTX_MAP_V_TILE: if(ReadFloat(&vt)) return TRUE; break;
		case RTX_MAP_U_ANGLE: if(ReadFloat(&ua)) return TRUE; break;
		case RTX_MAP_V_ANGLE: if(ReadFloat(&va)) return TRUE; break;
		case RTX_MAP_W_ANGLE: if(ReadFloat(&wa)) return TRUE; break;
		}
	}
}

BOOL RTXLoader::ReadMaterialProps()
{
	int id,dum,nrSub=0;
	DXC_MaterialInfo *pMi;
	aktMat->numSubMat=0;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_MATERIAL_NAME: if(ReadText(aktMat->Name)) return TRUE; break;
		case RTX_MATERIAL_DIFFUSE: if(ReadFVect3D((FVect3D*)&aktMat->diff)) return TRUE; 
			aktMat->diff.alpha=1; break;
		case RTX_MATERIAL_AMBIENT: if(ReadFVect3D((FVect3D*)&aktMat->ambi)) return TRUE; 
			aktMat->ambi.alpha=1; break;
		case RTX_MATERIAL_SPECULAR: if(ReadFVect3D((FVect3D*)&aktMat->spec)) return TRUE; 
			aktMat->spec.alpha=1; break;
		case RTX_MATERIAL_SHINE: if(ReadFloat(&aktMat->strength)) return TRUE; break;
		case RTX_MAT_GLOSSINESS: if(ReadFloat(&aktMat->gloss)) return TRUE; break;
		case RTX_MAP_DIFFUSE: Goto('{'); if(ReadMapDiffuse()) return TRUE; break;
		case RTX_MAT_NUMSUB: ReadInt(&aktMat->numSubMat); break;
		case RTX_MAT_SUB: ReadInt(&dum); if(dum!=nrSub) DXCError("Submaterial in wrong order!");
			nrSub++; pMi=aktMat; dum=aktMat->SubMat.New(); aktMat=pMi->SubMat.Ptr(dum); 
			Goto('{'); ReadMaterialProps(); aktMat=pMi; break;
		case RTX_MAT_TRANSPARENCY: ReadFloat(&aktMat->transpar); break;
		case RTX_MAT_2SIDED: 
			if(ReadInt(&dum)) { aktMat->Is2Sided=TRUE; break; }
			else
				if(dum==0) aktMat->Is2Sided=FALSE;
				else aktMat->Is2Sided=TRUE;			
		}
	}
}

BOOL RTXLoader::ReadMaterials()
{
	int id,dum,aktNr=0;
	int numMat;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		case RTX_MATERIAL_COUNT: ReadInt(&numMat); break;
		case RTX_MATERIAL: 
			dum=MatList.New(); aktMat=MatList.Ptr(dum); 
			ReadInt(&dum); if(dum!=aktNr) DXCError("Material in wrong order!");
			Goto('{'); strcpy(aktMat->DiffTextFile,""); ReadMaterialProps(); aktNr++; break;
		}
	}
}

BOOL RTXLoader::ReadScene()
{
	int id;
	FVect3D dum;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_FIRSTFRAME: ReadInt(&FirstFrame); break;
		case RTX_LASTFRAME: ReadInt(&LastFrame); break;
		case RTX_TICKSPERFRAME: ReadFloat(&TicksPerFrame); break;
		case RTX_FRAMESPEED: ReadFloat(&FrameSpeed); break;
		case RTX_BACKGROUND_STATIC: ReadFVect3D(&dum);
			m_Ani->Backgrd.red=(BYTE)dum.x*255;
			m_Ani->Backgrd.green=(BYTE)dum.y*255;
			m_Ani->Backgrd.blue=(BYTE)dum.z*255; break;
		case RTX_AMBIENT_STATIC: ReadFVect3D(&dum);
			m_Ani->AmbientCol.red=(BYTE)dum.x*255;
			m_Ani->AmbientCol.green=(BYTE)dum.y*255;
			m_Ani->AmbientCol.blue=(BYTE)dum.z*255; break;
		case RTX_SCENE_ENVMAP: Goto('{'); ReadEnvMap(); break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		}
	}
}

BOOL RTXLoader::ReadTM(DXC_TM_Info *tm)
{
	int id,nr;
	float f;
	FVect3D v;
	D3DXMATRIX RotMat,ScaleMat,PosMat,Mat;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_TM_ROW0: ReadFVect3D(&v); tm->mat(0,0)=v.x; tm->mat(0,1)=v.y; tm->mat(0,2)=v.z; tm->mat(0,3)=0; break;
		case RTX_TM_ROW1: ReadFVect3D(&v); tm->mat(1,0)=v.x; tm->mat(1,1)=v.y; tm->mat(1,2)=v.z; tm->mat(1,3)=0; break;
		case RTX_TM_ROW2: ReadFVect3D(&v); tm->mat(2,0)=v.x; tm->mat(2,1)=v.y; tm->mat(2,2)=v.z; tm->mat(2,3)=0; break;
		case RTX_TM_ROW3: ReadFVect3D(&v); tm->mat(3,0)=v.x; tm->mat(3,1)=v.y; tm->mat(3,2)=v.z; tm->mat(3,3)=1; break;
		case RTX_TM_POS: ReadFVect3D(&tm->pos); break;
		case RTX_TM_SCALE: ReadFVect3D(&tm->scale); break;
		case RTX_TM_ROTAXIS: ReadFVect3D(&tm->axis); tm->axis*=DXC_ROTATION_SWAP;
			break;
		case RTX_TM_ROTANGLE: ReadFloat(&tm->angle);  
			break;
		case -1: case -2: 
			D3DXMatrixRotationAxis(&RotMat,(D3DXVECTOR3*)&tm->axis,tm->angle);
			D3DXMatrixScaling(&ScaleMat,tm->scale.x,tm->scale.y,tm->scale.z);
			D3DXMatrixTranslation(&PosMat,tm->pos.x,tm->pos.y,tm->pos.z);
			Mat=ScaleMat*RotMat*PosMat;
			tm->mirror=FVect3D(1,1,1);

			// scaling is always positive, you need to get the 'mirror value' from
			// the transformation matrix
			nr=absmax3(Mat(0,0),Mat(1,0),Mat(2,0));
			f=tm->mat(nr,0);
			if(f==0) { 
				DXCWarn("Determinant in TM is zero."); return FALSE; }
			if(Mat(nr,0)/tm->mat(nr,0)<0) tm->mirror.x=-1;
			nr=absmax3(Mat(0,1),Mat(1,1),Mat(2,1));
			if(tm->mat(nr,1)==0) { DXCWarn("Determinant in TM is zero."); return FALSE; }
			if(Mat(nr,1)/tm->mat(nr,1)<0) tm->mirror.y=-1;
			nr=absmax3(Mat(0,2),Mat(1,2),Mat(2,2));
			if(tm->mat(nr,2)==0) { DXCWarn("Determinant in TM is zero."); return FALSE; }
			if(Mat(nr,2)/tm->mat(nr,2)<0) tm->mirror.z=-1;
						
			if(id==-2) return FALSE; else return TRUE;
		case 0: break;
		}
	}
}	

BOOL RTXLoader::ReadCamSettings()
{
	int id;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_CAM_FOV: ReadFloat(&aktCam->FOV); break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;

		}
	}
}
		
BOOL RTXLoader::ReadCamera()
{
	int id;
	
	id=m_Ani->pCam.New();
	aktCam=m_Ani->pCam.Ptr(id);
	aktCam->ObjectType=DXC_3DOBJ_CAMERA;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_NODE_TM: Goto('{'); ReadTM(&aktCam->m_TM_Info); break;
		case RTX_CAM_SETTINGS: Goto('{'); ReadCamSettings(); break;
		case RTX_NODE_PARENT: ReadText(aktCam->ParentName); break;
		case RTX_NODE_NAME: ReadText(aktCam->Name); break;
		case RTX_CAM_TARGET: ReadText(aktCam->TargetName); break;
		case RTX_ANIMATION: Goto('{'); aktObj=aktCam; ReadAnimation(); break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		}
	}
}

BOOL RTXLoader::ReadObject()
{
	int id;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_NODE_NAME: ReadText(aktHelper->Name); break;
		case RTX_ANIMATION: Goto('{'); aktObj=aktHelper; ReadAnimation(); break;
		case RTX_NODE_PARENT: ReadText(aktHelper->ParentName); break;
		case RTX_NODE_TM: Goto('{'); ReadTM(&aktHelper->m_TM_Info); break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		}
	}
}

BOOL RTXLoader::ReadLightSettings()
{
	int id;
	FVect3D v;
	float f;
	DXC_ColorF col;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_LIGHT_COLOR: ReadFVect3D(&v); break;
		case RTX_LIGHT_INTENS: ReadFloat(&f); v*=f;
			col.red=v.x; col.green=v.y; col.blue=v.z;
			aktLight->light.Diffuse.r=col.red; 
			aktLight->light.Diffuse.g=col.green; 
			aktLight->light.Diffuse.b=col.blue; 
			aktLight->light.Diffuse.a=1; 
			aktLight->light.Specular.r=col.red;
			aktLight->light.Specular.g=col.green;
			aktLight->light.Specular.b=col.blue; 
			aktLight->light.Specular.a=1; 
			aktLight->light.Ambient.r=0;
			aktLight->light.Ambient.g=0;
			aktLight->light.Ambient.b=0;
			aktLight->light.Ambient.a=0;
			break;
		case RTX_LIGHT_ASPECT: ReadFloat(&f); if(f!=1) DXCWarn("Light aspect other than 1"); break;
		case RTX_LIGHT_HOTSPOT: ReadFloat(&f); aktLight->light.Theta=RAD(f); break;
		case RTX_LIGHT_FALLOFF: ReadFloat(&f); aktLight->light.Phi=RAD(f); break;
		case RTX_LIGHT_TDIST: ReadFloat(&f); aktLight->light.Range=LGHT_MAX_RANGE; break;
		case -2: return FALSE;
		case -1: return TRUE;
		case 0: break;
		}
	}
}
	
BOOL RTXLoader::ReadLight()
{
	int id,dum,i;
	char LightTyp[128];
	dum=m_Ani->pLight.New(); aktLight=m_Ani->pLight.Ptr(dum); 
	aktLight->ObjectType=DXC_3DOBJ_LIGHT;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case RTX_NODE_NAME: ReadText(aktLight->Name); break;
		case RTX_ANIMATION: Goto('{'); aktObj=aktLight; ReadAnimation(); break;
		case RTX_NODE_PARENT: ReadText(aktLight->ParentName); break;
		case RTX_NODE_TM: Goto('{'); ReadTM(&aktLight->m_TM_Info); break;
		case RTX_LIGHT_SETTINGS: Goto('{'); ReadLightSettings(); break;
		case RTX_USELIGHT: ReadInt(&i); 
			if(i!=0) aktLight->IsOn=TRUE;
			else aktLight->IsOn=FALSE; break;
		case RTX_LIGHT_TYPE: Skip(); ReadText2(LightTyp); break;
		case -2: 
			aktLight->light.Falloff=1;
			aktLight->light.Attenuation0=1;
			aktLight->light.Attenuation1=0;
			aktLight->light.Attenuation2=0;
			aktLight->bEmpty=FALSE;
			if(strcmp(LightTyp,"Target")==0) { 
				aktLight->light.Type=D3DLIGHT_SPOT;
				return FALSE;
			} else if(strcmp(LightTyp,"Directional")==0) {
				aktLight->light.Type=D3DLIGHT_DIRECTIONAL;
			} else if(strcmp(LightTyp,"Omni")!=0) DXCWarn("Unknown light type changed to omni");
			aktLight->light.Type=D3DLIGHT_POINT;
			return FALSE;
		case -1: return TRUE;
			
		case 0: break;
		}
	}
}


// Calculate the Materials to be used in DirectX 8
void RTXLoader::PrepareMaterial(DXC_MaterialInfo *mi)
{
	int i;
	mi->Mat.diff=mi->diff; mi->Mat.diff.alpha=1-mi->transpar;
	mi->Mat.ambi=mi->ambi; mi->Mat.ambi.alpha=1-mi->transpar;
	mi->Mat.spec=mi->spec*mi->strength; mi->Mat.spec.alpha=1-mi->transpar;
	mi->Mat.power=mi->gloss*100; 
	mi->Mat.emis=0; mi->Mat.Is2Sided=mi->Is2Sided;
	if(mi->DiffTextFile[0]!=0&&m_LoadTex==TRUE) {
		dgfx->LoadTexture(&mi->Tex,mi->DiffTextFile,0,DXC_STRETCH);
		mi->IsTexture=TRUE;
	} else {
		mi->Tex.SetDevice(dgfx->pDev);
		mi->IsTexture=FALSE;
	}
	for(i=0;i<mi->SubMat.GetSize();i++)
	{
		PrepareMaterial(mi->SubMat.Ptr(i));
	}
}

void RTXLoader::PrepareMaterials()
{
	int i;
	for(i=0;i<MatList.GetSize();i++)
	{
		PrepareMaterial(MatList.Ptr(i));
	}
}

BOOL RTXLoader::ReadMain()
{
	int id,dum;
	for(;;)
	{
		id=ReadNext();
		switch(id)
		{
		case -1: return TRUE;
		case 0: break;
		case RTX_SCENE: Goto('{'); if(ReadScene()) return FALSE; 
			TimeFactor=1.f/(TicksPerFrame*FrameSpeed); 
			TotalTime=((LastFrame-FirstFrame)/FrameSpeed); break;
		case RTX_MATERIAL_LIST: Goto('{'); if(ReadMaterials()) return FALSE; 
			PrepareMaterials(); 
			break;
		case RTX_GEOMOBJECT: Goto('{'); if(ReadGeomObject()) return FALSE; break;
		case RTX_BONE: 
			dum=m_Ani->pHelper.New(); aktHelper=m_Ani->pHelper.Ptr(dum); 
			aktHelper->ObjectType=DXC_3DOBJ_BONE;
			Goto('{'); if(ReadObject()) return FALSE; 
			break;
		case RTX_HELPER: case RTX_TARGETOBJ: 
			dum=m_Ani->pHelper.New(); aktHelper=m_Ani->pHelper.Ptr(dum); 
			aktHelper->ObjectType=DXC_3DOBJ_HELPER;
			Goto('{'); if(ReadObject()) return FALSE; 
			break;
		case RTX_GROUP: Goto('{'); if(!ReadMain()) return FALSE; break;
		case RTX_CAMERA: Goto('{'); if(ReadCamera()) return FALSE; break; 
		case RTX_LIGHT: Goto('{'); if(ReadLight()) return FALSE; break;
		default: break; 
		}
	}
}

//time scale needs to be adjusted in keys
void RTXLoader::AdjustKeys(DXC_3dObject *pO)
{
	DXC_PosKey *pPK;
	DXC_ScaleKey *pSK;
	DXC_RotKey *pRK;
	float dt;
	pPK=pO->FirstPosKey;
	while(pPK)
	{
		if(pPK->IntMode!=1) break;
		if(pPK->Prev!=0) {
			dt=pPK->time-pPK->Prev->time;
			pPK->Data.TangIn*=dt;
		}
		if(pPK->Next!=0) {
			dt=pPK->Next->time-pPK->time;
			pPK->Data.TangOut*=dt;
		}
		pPK=pPK->Next;
	}
	pPK=pO->FirstPosKey;
	while(pPK)
	{
		pPK->time*=TimeFactor;
		pPK=pPK->Next;
	}
	pSK=pO->FirstScaleKey;
	while(pSK)
	{
		if(pSK->IntMode!=1) break;
		if(pSK->Prev!=0) {
			dt=pSK->time-pSK->Prev->time;
			pSK->Data.TangIn*=dt;
		}
		if(pSK->Next!=0) {
			dt=pSK->Next->time-pSK->time;			
			pSK->Data.TangOut*=dt;
		}
		pSK=pSK->Next;
	}
	pSK=pO->FirstScaleKey;
	while(pSK)
	{
		pSK->time*=TimeFactor;
		pSK=pSK->Next;
	}
	pRK=pO->FirstRotKey;
	while(pRK)
	{
		pRK->time*=TimeFactor;
		pRK=pRK->Next;
	}
}

void RTXLoader::LoadStart()
{
	FirstFrame=0; LastFrame=100;
	FrameSpeed=30; TicksPerFrame=160;
	m_Ani->Free();
	m_fp=fopen(FileName,"rb");
	if(m_fp==0) DXCError("Can not open file %s",FileName);
	
	fseek(m_fp,0,SEEK_END);
	m_FileSize=ftell(m_fp);
	fseek(m_fp,0,SEEK_SET);
	m_curFilePos=0;
	if(m_bBufRead) {
		filebuf=new BYTE[m_FileSize];
		int anz=fread(filebuf,1,m_FileSize,m_fp);
		if(anz!=m_FileSize) 
		{ DXCError("Error reading from file %s at %i",FileName,anz); return; }
		fseek(m_fp,0,SEEK_SET);
	}
}

void RTXLoader::LoadEnd()
{
	fclose(m_fp);
	if(filebuf) delete[] filebuf;
	filebuf=NULL;
	MatList.Free();
}


//Info: DXC_3dScene is a container class with dynamically
//      allocated meshes (DXC_Mesh) helper objects (DXC_3dObject)
//      cameras (DXC_Camera) and lights (DXC_Light)

BOOL RTXLoader::Load(DXC_3dScene *ani, char fname[])
{
	int i;
	DXC_Material mat;
	DXC_Texture tex;
	DXC_3dSceneOptions opt;
	strcpy(FileName,fname); strcpy(ani->Name,fname);
	m_Ani=ani; m_bBufRead=TRUE; m_LoadTex=TRUE;
	LoadStart();
	if(!ReadMain())
		DXCWarn("Encountered unexpected end of file at pos %i",m_curFilePos);
	LoadEnd();
	if(m_Ani->pCam.GetSize()>0) m_Ani->SetCamera(0); 
	m_Ani->SetTotalTime(TotalTime);
	m_Ani->SetParentScene(m_Ani);
	opt.InverseX=TRUE; // 3ds MAX uses a right hand coord system
	opt.UseAmbient=TRUE;
	opt.ClearBg=TRUE;
	m_Ani->SetOptions(&opt);
	for(i=0;i<m_Ani->pMesh.GetSize();i++) {
		AdjustKeys(&m_Ani->pMesh(i));
	}
	for(i=0;i<m_Ani->pHelper.GetSize();i++) {
		AdjustKeys(&m_Ani->pHelper(i));
	}
	for(i=0;i<m_Ani->pCam.GetSize();i++) {
		AdjustKeys(&m_Ani->pCam(i));
	}
	for(i=0;i<m_Ani->pLight.GetSize();i++) {
		AdjustKeys(&m_Ani->pLight(i));
	}
	
	return TRUE;
}