/***********************************************************

This file is part of my 3d engine (DXC) and is not designed 
to be used in other applications. It is neither optimized 
nor does it cover all features of RTX. I can not guarantee
is functionality and I am not responsible for any malfunctions 
including possible damage to your computer system. Yet, I
hope it may be helpful to use parts of the code to load RTX 
files to other applications.
This source code my be distributed and used freely.

                           Michael.Schiestl@cluetec.com

**********************************************************/

#define DXC_RTX_H

#define RTX_SCENE              0x0001
#define RTX_BACKGROUND_STATIC  0x0011
#define RTX_AMBIENT_STATIC     0x0021
#define RTX_FIRSTFRAME         0x0031
#define RTX_LASTFRAME          0x0041
#define RTX_FRAMESPEED         0x0051
#define RTX_TICKSPERFRAME      0x0061
#define RTX_SCENE_ENVMAP       0x0071
#define RTX_MATERIAL_LIST      0x0002
#define RTX_MATERIAL_COUNT     0x0012
#define RTX_MATERIAL           0x0022
#define RTX_MATERIAL_NAME      0x0122
#define RTX_MAP_DIFFUSE        0x0222
#define RTX_BITMAP             0x1222
#define RTX_MAP_U_OFF		   0x2222
#define RTX_MAP_V_OFF		   0x3222
#define RTX_MAP_U_TILE		   0x4222
#define RTX_MAP_V_TILE		   0x5222
#define RTX_MAP_U_ANGLE		   0x6222
#define RTX_MAP_V_ANGLE		   0x7222
#define RTX_MAP_W_ANGLE		   0x8222
#define RTX_MATERIAL_DIFFUSE   0x0322
#define RTX_MATERIAL_AMBIENT   0x0422
#define RTX_MATERIAL_SPECULAR  0x0522
#define RTX_MATERIAL_SHINE     0x0622
#define RTX_MAT_GLOSSINESS     0x0722
#define RTX_MAT_NUMSUB         0x0822        
#define RTX_MAT_SUB            0x0922
#define RTX_MAT_TRANSPARENCY   0x0A22
#define RTX_MAT_2SIDED		   0x0B22
#define RTX_MATERIAL_REF       0x0032
#define RTX_GEOMOBJECT         0x0003
#define RTX_MESH               0x0023
#define RTX_MESH_NUMVERTEX     0x0123
#define RTX_MESH_NUMFACES      0x0223
#define RTX_MESH_VERTEX_LIST   0x0323
#define RTX_MESH_VERTEX        0x1323
#define RTX_MESH_FACE_LIST     0x0423
#define RTX_MESH_FACE          0x1423
#define RTX_MESH_SMOOTHING     0x2423
#define RTX_MESH_MTLID         0x3423
#define RTX_MESH_NUMTVERTEX    0x0523
#define RTX_MESH_TVERTLIST     0x0623
#define RTX_MESH_TVERT         0x1623
#define RTX_MESH_NUMTVFACES    0x0723
#define RTX_MESH_TFACELIST     0x0823
#define RTX_MESH_TFACE         0x1823
#define RTX_MESH_WEIGHTS       0x0923
#define RTX_MESH_INIT_TM       0x1923
#define RTX_BONE_WEIGHT        0x2923
#define RTX_BONE_NAME         0x12923
#define RTX_BONE_INIT_TM      0x22923
#define RTX_INDEX_WEIGHT      0x32923
#define RTX_GROUP              0x0004
#define RTX_ANIMATION          0x0005
#define RTX_CTRL_POS           0x0015
#define RTX_CTRL_POS_KEY       0x0115
#define RTX_CTRL_POS_KEY_BEZ   0x0215
#define RTX_CTRL_ROT           0x0025
#define RTX_CTRL_ROT_KEY       0x0125
#define RTX_CTRL_ROT_KEY_TCB   0x0225
#define RTX_CTRL_SCALE         0x0035
#define RTX_CTRL_SCALE_KEY     0x0135
#define RTX_CTRL_SCALE_KEY_BEZ 0x0235
#define RTX_CAMERA             0x0006
#define RTX_CAM_SETTINGS       0x0016
#define RTX_CAM_FOV            0x0116
#define RTX_CAM_TARGET         0x0026
#define RTX_BONE               0x0007
#define RTX_HELPER             0x0008
#define RTX_TARGETOBJ          0x0009
#define RTX_LIGHT              0x000A
#define RTX_LIGHT_SETTINGS     0x001A
#define RTX_LIGHT_COLOR        0x011A
#define RTX_LIGHT_INTENS       0x021A
#define RTX_LIGHT_ASPECT       0x031A
#define RTX_LIGHT_HOTSPOT      0x041A
#define RTX_LIGHT_FALLOFF      0x051A
#define RTX_LIGHT_TDIST        0x061A
#define RTX_USELIGHT           0x002A
#define RTX_LIGHT_TYPE         0x003A
#define RTX_NODE_TM            0x001F
#define RTX_TM_ROW0            0x011F
#define RTX_TM_ROW1            0x021F
#define RTX_TM_ROW2            0x031F
#define RTX_TM_ROW3            0x041F
#define RTX_TM_POS             0x051F
#define RTX_TM_ROTAXIS         0x061F
#define RTX_TM_ROTANGLE        0x071F
#define RTX_TM_SCALE           0x081F
#define RTX_NODE_NAME          0x002F
#define RTX_NODE_PARENT        0x003F
#define RTX_NODE_VISIBLE       0x004F

#define RTX_MAX_SMGGROUP 8
#define RTX_MAX_VERMULTI 128

// direction of rotation has to be reverted when using a left hand coord system
#define DXC_ROTATION_SWAP -1

struct FVect3D
{
	float x,y,z;
};

struct FVect2D
{
	float x,y;
};

template <typename Typ>
struct DXC_KeyBase {
	typedef DXC_KeyBase<Typ> MyType;
	float time;
	DXC_KeyBase *Next,*Prev;
	Typ Data;
	WORD IntMode;  //Interpolation: 0 linear, 1 Bezier, 2 TCB 
	DXC_KeyBase() {
		time=0; Next=NULL; Prev=NULL;
	}
	~DXC_KeyBase() {
		if(Next) delete Next;
	}
	DXC_KeyBase *Add(MyType *key) {
		if(time>key->time) {
			MyType *dum;
			dum=new MyType;
			*dum=*key;
			if(Prev) Prev->Next=dum;
			dum->Prev=Prev;
			dum->Next=this;
			Prev=dum;
			return dum;
		} else {
			if(Next) return Next->Add(key);
			else {
				Next=new MyType;
				*Next=*key;
				Next->Next=NULL;
				Next->Prev=this;
				return Next;
			}
		}
	}
	MyType *Get(float t)
	{
		if(Next==NULL)return this;
		else if(Next->time>t) return this;
			else return Next->Get(t);
	}
};

struct DXC_ScaleKeyData {
	FVect3D scale,TangIn,TangOut;
};

struct DXC_RotKeyData {
	FVect3D axis; //relativ
	float angle;
	float tens,cont,bias,easein,easeout;
	D3DXQUATERNION quart; //akkumuliert
};

struct DXC_PosKeyData {
	FVect3D pos,TangIn,TangOut;
};

struct DXC_TM_Info {
	D3DXMATRIX mat;
	FVect3D mirror;
	FVect3D pos,scale;
	FVect3D axis;
	float angle;
public:
	DXC_TM_Info() {
		D3DXMatrixIdentity(&mat);
		pos=FVect3D(0,0,0);
		scale=FVect3D(1,1,1);
		angle=0;
	}
};

typedef DXC_KeyBase<DXC_ScaleKeyData> DXC_ScaleKey;
typedef DXC_KeyBase<DXC_PosKeyData> DXC_PosKey;
typedef DXC_KeyBase<DXC_RotKeyData> DXC_RotKey;

struct DXC_BoneInfo {
	DXC_NameString Name;
	D3DXMATRIX m_BoneInitInv,m_MeshInit,m_MeshInitInv;
};

struct DXC_MaterialInfo
{
	char Name[256];
	char DiffTextFile[255];			// The texture file
	int numSubMat;
	//int MatID;
	DXC_ColorF diff;
	DXC_ColorF ambi;
	DXC_ColorF spec;
	float strength;
	float gloss;
	float transpar;
	BOOL IsTexture;
	BOOL Is2Sided;
	DynaMem<DXC_MaterialInfo,64> SubMat;
	DXC_Material Mat;
	DXC_Texture Tex;
	DXC_Matrix matr;
	DXC_MaterialInfo() {
		DiffTextFile[0]=0; Name[0]=0; IsTexture=FALSE; numSubMat=0; Is2Sided=FALSE;
		diff=0; ambi=0; spec=0; strength=0; gloss=0; transpar=0; matr.Identity();
	}
};

struct DXC_SmoothingGroupSet {
	BYTE SmGr[RTX_MAX_SMGGROUP]; //Smoothing Group
	short NrSmGr;
};

struct DXC_FaceInternData {
	
	DXC_SmoothingGroupSet sg;
	int MatId;  // Material SubID
	
};

struct DXC_FaceIntern : DXC_Face, DXC_FaceInternData {
	int w[3];
	int Attr;
};

struct DXC_Bone_Weight {
	int index;
	float weight;
};

struct DXC_Bone_Weight_Info {
	char BoneName[256];
	DXC_Bone_Weight *wght;
	DXC_Bone_Weight_Info() {
		BoneName[0]=0; wght=NULL;
	}
};

struct DXC_VertexIntern {
	FVect3D pos;
	float w[4];
	union {
		BYTE Bones[4];
		DWORD dwBone;
	};
	WORD VerId[RTX_MAX_VERMULTI];
	BYTE NumBone,NumVer;
};

struct DXC_VertexIntern2 {
	FVect3D pos;
	FVect2D t;
	float w[4];
	BYTE numTex;
	union {
		BYTE Bones[4];
		DWORD dwBone;
	};
	DXC_SmoothingGroupSet sg;
	short SubMat;
	short NrFaces;
	FVect3D NormalSum;
};

class RTXLoader : public DXC_Base
{
friend class DXC_Gfx;
friend class DXC_3dScene;
protected:
	DXC_Gfx *dgfx;
	BYTE *filebuf;
	BOOL m_bBufRead,m_LoadTex;
	float FrameSpeed,TicksPerFrame,TimeFactor;
	float TotalTime;
	int FirstFrame,LastFrame;
	int m_FileSize;
	int m_curFilePos;
	int visize;
	int aktMatID,aktVer2;
	FILE *m_fp;
	FVect3D *tvbuf;
	DXC_Face *tfbuf;
	DXC_VertexIntern *vbuf;
	DXC_FaceIntern *fbuf;
	DXC_VertexIntern2 *vbuf2;
	DXC_Mesh *aktMesh;
	DXC_3dObject *aktHelper;
	DXC_3dObject *aktObj;
	DXC_Light *aktLight;
	DynaMem<DXC_MaterialInfo,256> MatList;
	DXC_Texture *TexList;
	DXC_MaterialInfo *aktMat;
	DXC_3dScene *m_Ani;
	DXC_TM_Info m_MeshInitTM;
	DXC_Camera *aktCam;
	char FileName[256];
	int numVer,numFaces,numTVer,numTFaces;
	int ReadNext();
	int vread(void *ptr, int size);
	int vseek(int rel);
	BOOL Goto(char ch);
	BOOL ReadText2(char *txt);
	BOOL ReadNumber(char *txt);
	BOOL ReadText(char *txt);
	char ReadChar();
	char NextChar();
	BOOL ReadInt(int *i);
	BOOL ReadFloat(float *i);
	BOOL ReadFVect3D(FVect3D *v);
	BOOL ReadFVect2D(FVect2D *v);
	BOOL Skip();
	int Str2Id(char *str);
	BOOL Id2Str(char *str, int id);
	int absmax3(float v1, float v2, float v3);
		
	int RequestVertex(int nr, DXC_SmoothingGroupSet *sg, FVect3D normal, int SubMat);
	
	int AddToBoneList(DXC_BoneInfo *bi);
	void PrepareMaterials();
	void PrepareMaterial(DXC_MaterialInfo *mi);

	void AdjustKeys(DXC_3dObject *pO);
	int IntersectSmGr(DXC_SmoothingGroupSet *sg_in1, DXC_SmoothingGroupSet *sg_in2, DXC_SmoothingGroupSet *sg_out);

	BOOL ReadMain();
	BOOL ReadMesh();
	BOOL ReadMeshWeights();
	BOOL ReadBoneWeight();
	BOOL ReadAnimation();
	BOOL ReadEnvMap();
	BOOL ReadCtrlPos();
	BOOL ReadCtrlRot();
	BOOL ReadCtrlScale();
	BOOL ReadVertexList();
	BOOL ReadTVertexList();
	BOOL ReadFaceList();
	BOOL ReadTFaceList();
	BOOL ReadScene();
	BOOL ReadMaterials();
	BOOL ReadMaterialProps();
	BOOL ReadGeomObject();
	BOOL ReadMapDiffuse();
	BOOL ReadCamera();
	BOOL ReadCamSettings();
	BOOL ReadTM(DXC_TM_Info *tm);
	BOOL ReadObject();
	BOOL ReadLight();
	BOOL ReadLightSettings();
	void Set0();
	void SetGfx(DXC_Gfx *_gfx) { dgfx=_gfx; }
	float AdjustTile(float uv);
	void CalcMesh();
	void LoadStart();
	void LoadEnd();

public:
	void DXCError(char *err, ...);
	void DXCWarn(char *err, ...);
	BOOL Load(DXC_3dScene *ani, char fname[]);
	RTXLoader();
	~RTXLoader();
};
